import {TextDecoder, promisify} from 'util';
import {once} from 'events';
import * as stream from 'stream';

const streamFinished = promisify(stream.finished);
/**
 * @param {stream.Writable} outputStream
 */
export function chunkWriterAsync(outputStream) {
  outputStream.setDefaultEncoding('utf8');
  outputStream.addListener('error', (err) => {
    throw new Error(`Output stream error: ${err?.message ?? ''}`);
  });
  return {
    write: async (strChunk) => {
      if (!outputStream.write(strChunk, 'utf8')) {
        await once(outputStream, 'drain');
      }
    },
    done: async () => {
      outputStream.end();
      // Wait until flushed
      await once(outputStream, 'finish');
    },
  };
}

/**
 * @param {stream.Writable} stream
 * @param {string} lines
 */
export async function writeLines(stream, lines) {

  const chunkWriter = chunkWriterAsync(stream);
  // TODO: write large files in multiple chunks
  await chunkWriter.write(lines);
  await chunkWriter.done();
}

/**
 * @param {stream.Readable} stream
 */
export async function* parseLines(stream) {
  stream.addListener('error', (err) => {
    throw new Error(`Input stream error: ${err?.message ?? ''}`);
  });
  const utf8Decoder = new TextDecoder('utf-8');
  let outputBuffer = Buffer.from([]);
  let outputBufferPos = 0;
  let chunkCount = 0;
  for await (const chunk of stream) {
    chunkCount++;
    outputBuffer = outputBuffer.length > 0 ? Buffer.concat([outputBuffer, chunk]) : chunk;
    while (outputBufferPos < outputBuffer.length) {
      if (outputBuffer[outputBufferPos] === 0xA) { // newline
        const outputEndPos = (outputBufferPos > 0 && outputBuffer[outputBufferPos-1] === 0xD) ? // Drop CR before LF
          outputBufferPos - 1 : outputBufferPos;
        const nextChunk = outputBuffer.slice(0, outputEndPos);
        outputBuffer = outputBuffer.slice(outputBufferPos+1);
        outputBufferPos = 0;
        const nextChunkString = utf8Decoder.decode(nextChunk, {stream: false});
        yield nextChunkString;
      } else {
        outputBufferPos++;
      }
    }
  }
  if (outputBuffer.length > 0) { // Trailing string
    const nextChunkString = utf8Decoder.decode(outputBuffer, {stream: false});
    yield nextChunkString;
  }
}

/**
 * @param {AsyncGenerator<string, void, void>} lines
 */
export async function debugLogLines(lines, prefix = '') {
  let lineCount = 0;
  for await (const line of lines) {
    lineCount++;
    console.log(`${prefix}${line}`);
  }
}

/**
 * @param {AsyncGenerator<string, void, void>} lines
 */
export async function logOutLines(lines, prefix = '') {
  for await (const line of lines) {
    console.log(`${prefix}${line}`);
  }
}

/**
 * @param {AsyncGenerator<string, void, void>} lines
 */
export async function logErrLines(lines, prefix = '') {
  for await (const line of lines) {
    console.error(`${prefix}${line}`);
  }
}

/**
 * @param {AsyncGenerator<string, void, void>} lines
 */
export async function bufferLines(lines) {
  const output = [];
  for await (const line of lines) {
    output.push(line);
  }
  return output;
}

/**
 * @param {AsyncGenerator<string, void, void>} lines
 */
export async function bufferAndOutLines(lines, prefix = '') {
  const output = [];
  for await (const line of lines) {
    output.push(line);
    console.log(`${prefix}${line}`);
  }
  return output;
}

/**
 * @param {AsyncGenerator<string, void, void>} lines
 */
export async function bufferAndErrLines(lines, prefix = '') {
  const output = [];
  for await (const line of lines) {
    output.push(line);
    console.error(`${prefix}${line}`);
  }
  return output;
}
